.include "HDL50001_pcf.inc"

.alias SP  39
.alias DEC  0	/* Запрошена операция шифрования или расшифровывания */
.alias AA   1	/* Ai */
.alias BB   2	/* Bi */
.alias KI   3	/* Ki */
.alias BLK  4	/* Текущий номер блока */
.alias IDX  5	/* Номер раунда */

.global gost_encdec

.data

.align 8
Sblock:
    /* Определяем тестовые S-блоки в соответствии с ГОСТ Р 34.11-94 */
    .quad 0x35F7C1B6E08D29A4
    .quad 0x95701832AFD6C4BE
    .quad 0xB9067CFE243AD185
    .quad 0x352BC64EF9801AD7
    .quad 0x2B30E9A48DF517B6
    .quad 0xEFC95863D1270AB4
    .quad 0xC2867EA095F314BD
    .quad 0xC8B6E3294A750DF1

.bss

.align 8
Keys:
    .long

BufPtr:
    .long

.text

gost_encdec:
dec :=	rdl #SP, 4	/* operation */
key :=	rdl #SP, 8	/* key */
buf :=	rdl #SP, 12	/* buf */
blk :=	rdl #SP, 16	/* block count */

	setl #DEC, @dec
	wrl @key, Keys
	setl #BLK, @blk
	wrl @buf, BufPtr

	setl #32, Sblock

	jmp gost.next_block
complete

gost.next_block:
	getl #BLK
	subl @1, 1
	setl #BLK, @1
	je  @3, gost.return
	jne @4, gost.fetch_block
complete

gost.fetch_block:
	rdl BufPtr	/* указатель на буффер */
	addl @1, 4
	rdl @2		/* считывает два первых значения */
	rdl @2
	setl #AA, @2
	setl #BB, @2

	getl #DEC
	je  @1, gost_enc.prologue
	jne @2, gost_dec.prologue
complete

gost_enc.prologue:
	setl #IDX, 0
	jmp gost.load_key1_24
complete

gost_dec.prologue:
	setl #IDX, 31
	jmp gost.load_key25_32
complete

gost.load_key1_24:
kp :=	rdl Keys
	getl #IDX

	and  @1, 7		/* Ограничиваем i пределами [0..7] */
	slll @1, 2		/* Смещение до Ki */
	addl @kp, @1		/* Ki = kp + i*4 */
	rdl @1
	setl #KI, @1

	jmp gost.round
complete

gost.load_key25_32:
kp :=	rdl Keys
	getl #IDX

	and  @1, 7		/* Ограничиваем i пределами [7..0] */
	slll @1, 2		/* Смещение до Ki */
	addl @kp, 7 * 4		/* Пермещаем указатель к последнему элементу */
	subl @1, @2		/* Отсчитываем ключ с конца: Ki = kp - i*4 */
	rdl @1
	setl #KI, @1

	jmp gost.round
complete

gost.round:
	/* Сохраняем в коммутатор 8 значений Sblock */
s0 :=	rdq  #32
s1 :=	rdq  #32, 8
s2 :=	rdq  #32, 16
s3 :=	rdq  #32, 24
s4 :=	rdq  #32, 32
s5 :=	rdq  #32, 40
s6 :=	rdq  #32, 48
s7 :=	rdq  #32, 56

	getl #KI
	getl #AA
axk :=	xor  @1, @2		/* Сложение Ai + Ki по mod 2 */

	/* Разбиваем на 4-х битную подпоследовательность */
	slrl @axk, 4
	slrl @axk, 8
	slrl @axk, 12
	slrl @axk, 16
	slrl @axk, 20
	slrl @axk, 24
	slrl @axk, 28
	/* Получаем индекс для каждой подпоследовательности */
	and @axk, 0xf
	and @8, 0xf
	and @8, 0xf
	and @8, 0xf
	and @8, 0xf
	and @8, 0xf
	and @8, 0xf
	and @8, 0xf
	/* ...растягиваем, чтобы потом сдвигать вправо с кратностью по 4 бита */
	slll @8, 2
	slll @8, 2
	slll @8, 2
	slll @8, 2
	slll @8, 2
	slll @8, 2
	slll @8, 2
	slll @8, 2
	/* Получаем значение по индексу */
	slrq @s0, @8
	slrq @s1, @8
	slrq @s2, @8
	slrq @s3, @8
	slrq @s4, @8
	slrq @s5, @8
	slrq @s6, @8
	slrq @s7, @8
	and  @8, 0xf
	and  @8, 0xf
	and  @8, 0xf
	and  @8, 0xf
	and  @8, 0xf
	and  @8, 0xf
	and  @8, 0xf
	and  @8, 0xf
	/* Склеиваем подпоследовательности */
	slll @7, 4
	slll @7, 8
	slll @7, 12
	slll @7, 16
	slll @7, 20
	slll @7, 24
	slll @7, 28
	or   @15, @7
	or   @1, @7
	or   @1, @7
	or   @1, @7
	or   @1, @7
	or   @1, @7
	or   @1, @7

	roll @1, 11		/* Циклический сдвиг влево на 11 бит */
	getl #BB
	xor  @1, @2		/* Bi ^ f(Ai,Ki) */
	setl #BB, #AA		/* Bi+1 = Ai */
	setl #AA, @2		/* Ai+1 = Bi ^ f(Ai,Ki) */

	getl #DEC
	je @1, gost_enc.epilogue
	jne @2, gost_dec.epilogue
;	getl	0x1000			; 1: 0x2000 -> l 8192
;	wrl	@1, 0xc01f0304		; 2: @1 -> 1, 0xc01f0304 -> l 0xc01f0304
complete

gost_enc.epilogue:
	getl #IDX
	addl @1, 1
	setl #IDX, @1

	subl @2, 24
	subl @3, 32

	and  @2, 0x80000000
	jne  @1, gost.load_key1_24
	and  @3, 0x80000000
	xor  @1, @3
	jne  @1, gost.load_key25_32
	or   @5, @2
	je   @1, gost.save_results
complete

gost_dec.epilogue:
	getl #IDX
	subl @1, 1
	setl #IDX, @1

	subl @2, 24
	subl @3, 32

	and  @2, 0x80000000
	jne  @1, gost.load_key1_24
	and  @3, 0x80000000
	xor  @1, @3
	jne  @1, gost.load_key25_32
	and  @9, 0x80000000
	jne  @1, gost.save_results

;	getl	0x2000			; 1: 0x2000 -> l 8192
;	wrl	@1, 0xc01f0304		; 2: @1 -> 1, 0xc01f0304 -> l 0xc01f0304
complete

gost.save_results:
;	getl	0x3000			; 1: 0x2000 -> l 8192
;	wrl	@1, 0xc01f0304		; 2: @1 -> 1, 0xc01f0304 -> l 0xc01f0304
	rdl BufPtr
newp :=	addl @1, 4
	getl #BB
	getl #AA
	wrl @2, @4	/* Сохраняем новые значения в буфере */
	wrl @2, @4
	wrl @newp, BufPtr   /* Сохраняем новый указатель */
	jmp gost.next_block
complete

gost.return:
	getl #AA
	getl #BB
	rdl #SP
	jmp @1
complete
